# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.

import os
import warnings
from enum import Enum, auto
from pathlib import Path

from torch.cuda.memory import CUDAPluggableAllocator
from torch.utils.cpp_extension import CUDA_HOME, load_inline

from megatron.core.utils import is_torch_min_version

try:
    if is_torch_min_version("2.8.0"):
        from torch.cuda.memory import MemPool
    else:
        from torch.cuda import MemPool
    _has_mem_pool = True
except ImportError:
    _has_mem_pool = False


class CompilationState(Enum):
    """Enum to distinguish between unified memory (UVM) compilation states."""

    UNATTEMPTED = auto()  # Compilation has not been attempted.
    FAILURE = auto()  # Compilation attempted, but failed.
    SUCCESS = auto()  # Compilation attempted, and succeeded.


# Compilation vars.
_compilation_state = CompilationState.UNATTEMPTED
_alloc = None  # must remain global until process exit.
_mod = None  # must remain global until process exit.


class UnifiedMemoryUnsupportedError(Exception):
    """Unified memory is not supported on this system."""

    pass


def compile_allocator():
    """Attempt to compile UVM allocator."""

    global _compilation_state, _alloc, _mod

    if _compilation_state != CompilationState.UNATTEMPTED:
        return

    _mempool_c_src = r"""
    #include <cuda_runtime_api.h>
    #include <cstddef>

    #define EXPORT extern "C"

    EXPORT void* managed_malloc(size_t size, int device, void* stream) {
      (void)stream;
      int cur = -1;
      cudaGetDevice(&cur);
      if (device != cur && device >= 0) cudaSetDevice(device);

      // cudaMallocManaged allows for more memory to be allocated than the device memory size.
      // The cudaMemAttachGlobal flag makes the memory accessible from both host and device.
      void* ptr = nullptr;
      cudaError_t err = cudaMallocManaged(&ptr, (size_t)size, cudaMemAttachGlobal);
      if (err != cudaSuccess) return nullptr;

      if (device >= 0) {
        // cudaMemAdviseSetPreferredLocation sets the preferred location for the memory.
        // This is a hint that tries to prevent data from being migrated away from the device.
        cudaMemAdvise(ptr, (size_t)size, cudaMemAdviseSetPreferredLocation, device);
        // cudaMemAdviseSetAccessedBy ensures the memory always lives in the device's page table.
        // Even if the memory has to be migrated away from the device, it still does not page fault.
        // The CUDA docs claim that cudaMemAdviseSetPreferredLocation completely overrides this flag,
        // but there is no harm in adding this flag as well for future-proofing.
        cudaMemAdvise(ptr, (size_t)size, cudaMemAdviseSetAccessedBy, device);
      }
      return ptr;
    }

    EXPORT void managed_free(void* ptr, size_t size, int device, void* stream) {
      // Memory allocated with cudaMallocManaged should be released with cudaFree.
      (void)size; (void)device; (void)stream;
      if (ptr) cudaFree(ptr);
    }
    """

    # Build the .so upon import; this avoids issues.
    if _has_mem_pool:
        _extra_ldflags = ["-lcudart"]
        if CUDA_HOME:
            _cuda_lib = os.path.join(CUDA_HOME, "lib64")
            if os.path.isdir(_cuda_lib):
                _extra_ldflags = [f"-L{_cuda_lib}", "-lcudart"]
        try:
            _mod = load_inline(
                name="managed_alloc_runtime",
                cpp_sources=[_mempool_c_src],
                functions=[],
                with_cuda=True,
                extra_ldflags=_extra_ldflags,
                verbose=False,
            )
            _so_path = Path(_mod.__file__).as_posix()
            _alloc = CUDAPluggableAllocator(_so_path, "managed_malloc", "managed_free").allocator()
            _compilation_state = CompilationState.SUCCESS
        except (RuntimeError, ImportError, OSError) as e:
            warnings.warn(f"Failed to create unified memory mempool: '{e}'.")
            _compilation_state = CompilationState.FAILURE


def create_unified_mempool() -> "MemPool":
    """Create a unified memory mempool using CUDA managed memory.

    Returns:
        (MemPool) Unified memory mempool.
    """

    # Attempt to compile allocator.
    compile_allocator()

    # Return mempool.
    if _compilation_state != CompilationState.SUCCESS:
        raise UnifiedMemoryUnsupportedError()
    else:
        return MemPool(allocator=_alloc)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.

"""Retro's cross attention modules for the encoder block."""

from functools import partial
from typing import Callable, List, Optional, Tuple, Type

import torch
from torch import Tensor

from megatron.core.fusions.fused_bias_dropout import get_bias_dropout_add
from megatron.core.inference.contexts import BaseInferenceContext
from megatron.core.models.retro.base_attention import BaseRetroCrossAttention
from megatron.core.models.retro.config import RetroConfig
from megatron.core.models.retro.utils import get_all_true_mask
from megatron.core.transformer.module import MegatronModule
from megatron.core.utils import deprecate_inference_params


class RetroEncoderCrossAttention(BaseRetroCrossAttention):
    """Retro encoder's cross attention operator.

    See this paper for more details: https://arxiv.org/abs/2112.04426.
    Neighboring chunks are retrieved from the chunk database, encoded, and
    used by the decoder layers for chunked cross attention.

    Args:
        config (RetroConfig): Retro config.
        submodules (CrossAttentionSubmodules): Cross attention submodules.
        layer_number (int): Layer number within transformer block.
        attn_mask_type (AttnMaskType): Mask type ('causal' or 'padding').
    """

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: Tensor,
        key_value_states: Tensor = None,
        inference_context: BaseInferenceContext = None,
        # rotary_pos_emb: Tensor = None, # unsupported for retro.
        *,
        inference_params: Optional[BaseInferenceContext] = None,
    ) -> List[Tuple[Tensor, Optional[Tensor], Tensor]]:
        """Cross attention for Retro encoder.

        Notation:
            ns : Sequence length.
            bs : Batch size.
            d  : Hidden size.
            l  : Number of chunks per sample (i.e., seq_length/chunk_length).
            k  : Number of neighbors.
            r  : Number of retrieved tokens (neighbors + continuation).

        Args:
            hidden_states (Tensor): Transformer layer hidden states.
            attention_mask (Tensor): Attention mask.
            key_value_states (Tensor): Neighbor embeddings.
            inference_context (BaseInferenceContext): Inference context.

        Returns:
            List of tuples, where each tuple is (attention_output, attention_bias, residual).
        """

        inference_context = deprecate_inference_params(inference_context, inference_params)

        # Input shape. [ r, bs*l*k, d ]
        ns, bs, d = hidden_states.shape

        # Reshape sequence into neighboring chunks.
        # - hidden_states:   [ r, bs*l*k, d ]
        # - chunked_outputs: [ r, bs*l, k, d ]
        chunked_outputs = hidden_states.reshape(
            self.retro_retrieved_length, -1, self.retro_num_neighbors, d
        )

        # flash attn: [ b, h, sq, sk ]
        # fused attn: [ b, 1, 1, sq ]
        chunked_output_mask = get_all_true_mask(
            size=(1, 1, chunked_outputs.shape[0], key_value_states.shape[0]),
            device=chunked_outputs.device,
        )

        # Per-chunk attention.
        attention_output_tuples = []
        for k in range(self.retro_num_neighbors):

            # Attend to current neighboring chunks.
            # - chunked_output:   [ r, bs*l, d ]
            # - key_value_states: [ m, bs*l, d ]
            # - attention_output: [ r, bs*l, d ]
            # - attention_bias:   [ d ]
            chunked_output = chunked_outputs[:, :, k].contiguous()
            attention_output, attention_bias = self.attn(
                hidden_states=chunked_output,  # Q (neighbor embedding)
                attention_mask=chunked_output_mask,
                key_value_states=key_value_states,  # K, V (hidden act)
            )

            # Residual connection. [ r, bs*l, d ]
            residual = chunked_output

            # Collect tensors.
            attention_output_tuples.append((attention_output, attention_bias, residual))

        # Output. (List[Tuple[( [ r, bs*l, d ], [ d ] )]])
        return attention_output_tuples


class RetroEncoderBiasDropoutAdd(MegatronModule):
    """Retro encoder's bias-dropout-add operator.

    This operator applies bias-dropout-add individually on each neighboring
    chunk that is retrieved from the chunk database.

    Args:
        config (RetroConfig): Retro config.
    """

    def __init__(self, config: RetroConfig):
        super().__init__(config=config)
        self.retro_num_neighbors = config.retro_num_neighbors

    @classmethod
    def _forward(
        cls,
        x_with_bias: List[Tuple[Tensor, Optional[Tensor], Tensor]],
        residual: Tensor,
        prob: float,
        retro_num_neighbors: int,
        bias_dropout_add: Callable,
    ) -> Tensor:
        """Per-chunk bias-dropout-add.

        Args:
            x_with_bias (dict): Attention output and bias tuple.
            residual (Tensor): Transformer layer residual.
            prob (float): Dropout probability.
            retro_num_neighbors (int): Number of retrieved neighbor chunks (e.g., 2).
            bias_dropout_add (Callable): Bias-dropout-add function.

        Returns:
            Output of bias-dropout-add.
        """

        # Re-enable torch grad to enable fused optimization.
        with torch.enable_grad():

            # Per-neighbor bias-dropout-add.
            # - attention_output: [ r, bs*l, d ]
            # - attention_bias:   [ d ]
            # - residual:         [ r, bs*l, d ]
            # - output:           [ r, bs*l, d ]
            outputs = [
                bias_dropout_add(
                    (
                        attention_output,
                        None if attention_bias is None else attention_bias.expand_as(residual),
                    ),
                    residual,
                    prob,
                )
                for attention_output, attention_bias, residual in x_with_bias
            ]

        # Concatenate outputs (to shape [r, k*bs*l, d]; see notation above).
        r, _, d = outputs[0].shape
        output = torch.stack(outputs, dim=1).reshape(r, -1, d)

        # Output. [ r, k*bs*l, d ]
        return output

    def forward(self, training: bool, fused: bool) -> partial:
        """Retro decoder bias-dropout-add.

        Args:
            training (bool): If training, then apply dropout.
            fused (bool): Fuse bias-dropout-add.

        Returns:
            A partial function for performing bias-dropout-add.
        """
        return partial(
            self._forward,
            retro_num_neighbors=self.retro_num_neighbors,
            bias_dropout_add=get_bias_dropout_add(training, fused),
        )


class RetroEncoderLayerNorm(MegatronModule):
    """Retro encoder's layernorm operator.

    This operator applies layernorm individually on each neighboring chunk that
    is retrieved from the chunk database, and then concatenates the chunks into
    a single tensor.

    Args:
        config (RetroConfig): Retro config.
        submodules (Type): Layer norm class. (Named 'submodules' to fit external interface.)
    """

    def __init__(self, config: RetroConfig, submodules: Type, **kwargs: dict):
        super().__init__(config=config)
        norm_class = submodules
        self.norm = norm_class(config=config, **kwargs)
        self.retro_num_neighbors = config.retro_num_neighbors

    def forward(self, input: Tensor) -> Tensor:
        """Per-chunk layer norm.

        Args:
            input (Tensor): Input chunks, concatenated into a single tensor.

        Returns:
            Output of the layer norm.
        """

        # Input shape: [ r, k*bs*l, d ]. (see notation above in attention module)

        # Split input into 'num_neighbors' tensors.
        chunk_size = input.shape[1] // self.retro_num_neighbors
        inputs = torch.split(input, chunk_size, dim=1)

        # Norm.
        outputs = [self.norm(inp.contiguous()) for inp in inputs]

        # Concatenate layer norms (to shape [r, k*bs*l, d]; see notation above).
        r, _, d = inputs[0].shape
        output = torch.stack(outputs, dim=1).reshape(r, -1, d)

        # Output. [ r, k*bs*l, d ]
        return output



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.

import os

import torch


def get_config_path(project_dir: str) -> str:
    """Config copy stored within retro project dir."""
    return os.path.join(project_dir, "config.json")


def get_gpt_data_dir(project_dir: str) -> str:
    """Get project-relative directory of GPT bin/idx datasets."""
    return os.path.join(project_dir, "data")


# ** Note ** : Retro's compatibility between cross attention and Flash/Fused
#   Attention is currently a work in progress. We default to returning None for
#   now.
# def get_all_true_mask(size, device):
#     return torch.full(size=size, fill_value=True, dtype=torch.bool, device=device)
def get_all_true_mask(size, device):
    return None



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.

import logging
import os
import shutil
from typing import Optional

import torch
from torch.hub import _get_torch_home

try:
    import wget

    HAVE_WGET = True
except ModuleNotFoundError:
    HAVE_WGET = False

from .huggingface_tokenizer import HuggingFaceTokenizer

logger = logging.getLogger(__name__)
torch_home = _get_torch_home()

if not isinstance(torch_home, str):
    logger.info("Torch home not found, caching megatron in cwd")
    torch_home = os.getcwd()

MEGATRON_CACHE = os.path.join(torch_home, "megatron")

MEGATRON_CONFIG_MAP = {
    "BertWordPieceLowerCase": {
        "checkpoint": "https://api.ngc.nvidia.com/v2/models/nvidia/megatron_bert_345m/versions/v0.0/files/release/mp_rank_00/model_optim_rng.pt",  # pylint: disable=line-too-long
        "vocab": "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-vocab.txt",  # pylint: disable=line-too-long
        "tokenizer_name": "bert-large-uncased",
    },
    "BertWordPieceCase": {
        "checkpoint": "https://api.ngc.nvidia.com/v2/models/nvidia/megatron_bert_345m/versions/v0.1_cased/files/release/mp_rank_00/model_optim_rng.pt",  # pylint: disable=line-too-long
        "vocab": "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt",
        "tokenizer_name": "bert-large-cased",
    },
    "GPT2BPETokenizer": {
        "vocab": "https://s3.amazonaws.com/models.huggingface.co/bert/gpt2-vocab.json",
        "merges_file": "https://s3.amazonaws.com/models.huggingface.co/bert/gpt2-merges.txt",
        "tokenizer_name": "gpt2",
    },
    "megatron-gpt-345m": {
        "vocab": "https://s3.amazonaws.com/models.huggingface.co/bert/gpt2-vocab.json",
        "merges_file": "https://s3.amazonaws.com/models.huggingface.co/bert/gpt2-merges.txt",
        "tokenizer_name": "gpt2",
    },
    "megatron-bert-345m-uncased": {
        "checkpoint": "https://api.ngc.nvidia.com/v2/models/nvidia/megatron_bert_345m/versions/v0.0/files/release/mp_rank_00/model_optim_rng.pt",  # pylint: disable=line-too-long
        "vocab": "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-vocab.txt",
        "tokenizer_name": "bert-large-uncased",
    },
    "megatron-bert-345m-cased": {
        "checkpoint": "https://api.ngc.nvidia.com/v2/models/nvidia/megatron_bert_345m/versions/v0.1_cased/files/release/mp_rank_00/model_optim_rng.pt",  # pylint: disable=line-too-long
        "vocab": "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt",
        "tokenizer_name": "bert-large-cased",
    },
    "megatron-bert-uncased": {
        "vocab": "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-vocab.txt",  # pylint: disable=line-too-long
        "tokenizer_name": "bert-large-uncased",
    },
    "megatron-bert-cased": {
        "vocab": "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt",
        "tokenizer_name": "bert-large-cased",
    },
    "biomegatron-bert-345m-uncased": {
        "vocab": "https://api.ngc.nvidia.com/v2/models/nvidia/biomegatron345muncased/versions/0/files/vocab.txt",  # pylint: disable=line-too-long
        "tokenizer_name": "bert-large-uncased",
    },
    "biomegatron-bert-345m-cased": {
        "vocab": "https://api.ngc.nvidia.com/v2/models/nvidia/biomegatron345mcased/versions/0/files/vocab.txt",  # pylint: disable=line-too-long
        "tokenizer_name": "bert-large-cased",
    },
}


class MegatronHFTokenizer(HuggingFaceTokenizer):
    """ """

    def __init__(
        self,
        tokenizer_path: str,
        vocab_file: Optional[str] = None,
        merges_file: Optional[str] = None,
        **kwargs,
    ) -> None:
        if tokenizer_path in MEGATRON_CONFIG_MAP.keys():
            tokenizer_name = tokenizer_path
        else:
            raise ValueError(
                f"The name of the tokenizer is incorrect. \
            Please see the list of available models: {self._get_available_models_list()}."
            )

        vocab_file = self._get_vocab_file(tokenizer_name, vocab_file)
        merges_file = self._get_merges_file(tokenizer_name, vocab_file)
        tokenizer_path = MEGATRON_CONFIG_MAP[tokenizer_name]["tokenizer_name"]
        super().__init__(tokenizer_path, vocab_file, merges_file, **kwargs)

    def _get_vocab_file(self, tokenizer_name: str, vocab_file: str = None) -> str:
        """
        Gets vocabulary file from cache or downloads it.

        Args:
            tokenizer_name (str): pretrained model name.
            vocab_file (str): path to the vocab file.

        Returns:
            path: path to the vocab file
        """

        if not vocab_file:
            url = MEGATRON_CONFIG_MAP[tokenizer_name]["vocab"]

            path = os.path.join(MEGATRON_CACHE, tokenizer_name + "_vocab")
            vocab_file = self._download(path, url)

        return vocab_file

    def _get_merges_file(self, tokenizer_name: str, merges_file: str = None) -> str:
        """
        Gets merge file from cache or downloads it.

        Args:
            tokenizer_name (str): pretrained model name.
            merges_file (str): path to the merges file.

        Returns:
            path: path to the vocab file.
        """

        if not merges_file:
            if 'gpt' not in tokenizer_name.lower():
                return None
            url = MEGATRON_CONFIG_MAP[tokenizer_name]["merges_file"]

            path = os.path.join(MEGATRON_CACHE, tokenizer_name + "_merges")
            merges_file = self._download(path, url)

        return merges_file

    def _get_available_models_list(self) -> list:
        """Returns a list of available megatron tokenizers."""

        return list(MEGATRON_CONFIG_MAP.keys())

    def _download(self, path: str, url: str):
        """
        Gets a file from cache or downloads it

        Args:
            path: path to the file in cache
            url: url to the file
        Returns:
            path: path to the file in cache
        """

        if url is None:
            return None

        if (
            not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0
        ) and not os.path.exists(path):
            os.makedirs(MEGATRON_CACHE, exist_ok=True)
            logging.info(f"Downloading from {url} to {path}")
            if HAVE_WGET:
                downloaded_path = wget.download(url)
            else:
                raise ModuleNotFoundError("wget library should be isntalled.")
            if not os.path.exists(downloaded_path):
                raise FileNotFoundError(f"Downloaded file not found: {downloaded_path}")
            shutil.move(downloaded_path, path)
        # wait until the master process downloads the file and writes it to the cache dir
        if torch.distributed.is_initialized():
            torch.distributed.barrier()

        return path



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.

"""Container class for GPT and Bert tokenizers."""

from dataclasses import dataclass

from megatron.core.tokenizers import MegatronTokenizerBase


@dataclass
class RetroTokenizers:
    """Container class for GPT and Bert tokenizers."""

    gpt: MegatronTokenizerBase = None
    bert: MegatronTokenizerBase = None



# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.

from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import numpy

from megatron.core.datasets.indexed_dataset import IndexedDataset
from megatron.core.datasets.masked_dataset import (
    MaskedWordPieceDataset,
    MaskedWordPieceDatasetConfig,
)
from megatron.core.datasets.utils import Split


@dataclass
class BERTMaskedWordPieceDatasetConfig(MaskedWordPieceDatasetConfig):
    """Configuration object for Megatron Core BERT WordPiece datasets"""

    classification_head: bool = None
    """Option to perform the next sequence prediction during sampling"""

    def __post_init__(self) -> None:
        """Do asserts and set fields post init"""
        super().__post_init__()

        assert self.classification_head is not None


class BERTMaskedWordPieceDataset(MaskedWordPieceDataset):
    """The BERT dataset that assumes WordPiece tokenization

    Args:
        indexed_dataset (IndexedDataset): The IndexedDataset around which
            to build the MegatronDataset
        dataset_path (str): The real path on disk to the dataset, for bookkeeping
        indexed_indices (numpy.ndarray): The set of the documents indices to expose
        num_samples (Optional[int]): The number of samples to draw from the indexed dataset.
            When None, build as many samples as correspond to one epoch.
        index_split (Split): The indexed_indices Split
        config (BERTMaskedWordPieceDatasetConfig): The config
    """

    def __init__(
        self,
        indexed_dataset: IndexedDataset,
        dataset_path: str,
        indexed_indices: numpy.ndarray,
        num_samples: Optional[int],
        index_split: Split,
        config: BERTMaskedWordPieceDatasetConfig,
    ) -> None:
        super().__init__(
            indexed_dataset, dataset_path, indexed_indices, num_samples, index_split, config
        )

        self.token_lookup = list(self.config.tokenizer.inv_vocab.keys())
        # Account for the single <cls> and two <sep> token ids
        self.sample_index = self._build_sample_index(
            self.config.sequence_length - 3, 2 if self.config.classification_head else 1
        )

    @staticmethod
    def _key_config_attributes() -> List[str]:
        """Inherited method implementation

        Returns:
            List[str]: The key config attributes
        """
        return super(
            BERTMaskedWordPieceDataset, BERTMaskedWordPieceDataset
        )._key_config_attributes() + ["classification_head"]

    def __getitem__(self, idx: int) -> Dict[str, Union[int, numpy.ndarray]]:
        """Abstract method implementation

        Args:
            idx (int): The index into the dataset

        Returns:
            Dict[str, Union[int, numpy.ndarray]]: The
        """

        idx_beg, idx_end, target_sequence_length = self.sample_index[idx]
        sample = [self.dataset[i] for i in range(idx_beg, idx_end)]
        numpy_random_state = numpy.random.RandomState(seed=(self.config.random_seed + idx) % 2**32)

        assert target_sequence_length <= self.config.sequence_length

        # Split the sample into contiguous subsegments A and B
        pivot = len(sample)
        is_next_random = False
        if self.config.classification_head:
            assert len(sample) > 1, "the sample must contain at least two sentences"
            pivot = 1
            if len(sample) >= 3:
                pivot = numpy_random_state.randint(low=1, high=len(sample))
            is_next_random = numpy_random_state.random() < 0.5
        split_A = []
        for sample_a in sample[:pivot]:
            split_A.extend(sample_a)
        split_B = []
        for sample_b in sample[pivot:]:
            split_B.extend(sample_b)
        if is_next_random:
            split_A, split_B = split_B, split_A

        # Trim the subsegments from either end to a desired joint length
        length_A = len(split_A)
        length_B = len(split_B)
        if length_A + length_B <= target_sequence_length:
            truncated = False
        else:
            while length_A + length_B > target_sequence_length:
                split = split_A if length_A > length_B else split_B
                if numpy_random_state.random() < 0.5:
                    del split[0]
                else:
                    del split[-1]
                length_A = len(split_A)
                length_B = len(split_B)
            truncated = True

        # Merge the subsegments and create the token assignment labels
        tokens = [self.config.tokenizer.cls, *split_A, self.config.tokenizer.sep]
        assignments = [0 for _ in range(1 + len(split_A) + 1)]
        if split_B:
            tokens += [*split_B, self.config.tokenizer.sep]
            assignments += [1 for _ in range(len(split_B) + 1)]

        # Masking
        tokens, masked_positions, masked_labels, _, _ = self._create_masked_lm_predictions(
            tokens, target_sequence_length, numpy_random_state
        )

        # Pad the sequences and convert to NumPy
        length_toks = len(tokens)
        length_pads = self.config.sequence_length - length_toks
        assert length_pads >= 0

        tokens = numpy.array(tokens, dtype=numpy.int64)
        tokens = numpy.pad(tokens, (0, length_pads), constant_values=self._pad_token_id)

        assignments = numpy.array(assignments, dtype=numpy.int64)
        assignments = numpy.pad(assignments, (0, length_pads), constant_values=self._pad_token_id)

        # Get the padding mask
        mask_pads = numpy.ones(self.config.sequence_length, dtype=numpy.int64)
        mask_pads[tokens == self._pad_token_id] = self._pad_token_id

        # Mask the labels
        labels = numpy.zeros(self.config.sequence_length, dtype=numpy.int64) - 1
        labels[masked_positions] = masked_labels

        # Get the loss mask
        mask_loss = numpy.zeros(self.config.sequence_length, dtype=numpy.int64)
        mask_loss[masked_positions] = 1

        # For padded sequences, ensure the embedding layer can map the token ID
        tokens[tokens == self._pad_token_id] = 0
        labels[labels == self._pad_token_id] = 0

        return {
            "text": tokens,
            "types": assignments,
            "labels": labels,
            "is_random": int(is_next_random),
            "padding_mask": mask_pads,
            "loss_mask": mask_loss,
            "truncated": int(truncated),
        }

    def _get_token_mask(self, numpy_random_state: numpy.random.RandomState) -> Optional[int]:
        """Abstract method implementation

        80% of the time, replace the token id with mask token id. 10% of the time, replace token id
        with a random token id from the vocabulary. 10% of the time, do nothing.

        Args:
            numpy_random_state (RandomState): The NumPy random state

        Returns:
            Optional[int]: The replacement token id or None
        """
        if numpy_random_state.random() < 0.8:
            return self.config.tokenizer.mask
        else:
            if numpy_random_state.random() >= 0.5:
                return self.token_lookup[numpy_random_state.randint(0, len(self.token_lookup))]
        return None
